"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const cluster_key_slot_1 = __importDefault(require("cluster-key-slot"));
const client_1 = __importDefault(require("./client"));
class RedisClusterSlots {
    #options;
    #nodeByUrl = new Map();
    #slots = [];
    constructor(options) {
        this.#options = options;
    }
    async connect() {
        for (const rootNode of this.#options.rootNodes) {
            try {
                await this.#discoverNodes(rootNode);
                return;
            }
            catch (err) {
                console.error(err);
                // this.emit('error', err);
            }
        }
        throw new Error('None of the root nodes is available');
    }
    async discover(startWith) {
        try {
            await this.#discoverNodes(startWith.options?.socket);
            return;
        }
        catch (err) {
            console.error(err);
            // this.emit('error', err);
        }
        for (const { client } of this.#nodeByUrl.values()) {
            if (client === startWith)
                continue;
            try {
                await this.#discoverNodes(client.options?.socket);
                return;
            }
            catch (err) {
                console.error(err);
                // this.emit('error', err);
            }
        }
        throw new Error('None of the cluster nodes is available');
    }
    async #discoverNodes(socketOptions) {
        const client = client_1.default.create({
            socket: socketOptions
        });
        await client.connect();
        try {
            await this.#reset(await client.clusterNodes());
        }
        finally {
            await client.disconnect(); // TODO: catch error from disconnect?
        }
    }
    async #reset(masters) {
        // Override this.#slots and add not existing clients to this.#clientByKey
        const promises = [], clientsInUse = new Set();
        for (const master of masters) {
            const slot = {
                master: this.#initiateClientForNode(master, false, clientsInUse, promises),
                replicas: this.#options.useReplicas ?
                    master.replicas.map(replica => this.#initiateClientForNode(replica, true, clientsInUse, promises)) :
                    [],
                clientIterator: undefined // will be initiated in use
            };
            for (const { from, to } of master.slots) {
                for (let i = from; i < to; i++) {
                    this.#slots[i] = slot;
                }
            }
        }
        // Remove unused clients from this.#clientBykey using clientsInUse
        for (const [url, { client }] of this.#nodeByUrl.entries()) {
            if (clientsInUse.has(url))
                continue;
            // TODO: ignore error from `.disconnect`?
            promises.push(client.disconnect());
            this.#nodeByUrl.delete(url);
        }
        await Promise.all(promises);
    }
    #initiateClientForNode(nodeData, readonly, clientsInUse, promises) {
        const url = `${nodeData.host}:${nodeData.port}`;
        clientsInUse.add(url);
        let node = this.#nodeByUrl.get(url);
        if (!node) {
            node = {
                id: nodeData.id,
                client: client_1.default.create({
                    socket: {
                        host: nodeData.host,
                        port: nodeData.port
                    },
                    readonly
                })
            };
            promises.push(node.client.connect());
            this.#nodeByUrl.set(url, node);
        }
        return node;
    }
    getSlotMaster(slot) {
        return this.#slots[slot].master;
    }
    *#slotClientIterator(slotNumber) {
        const slot = this.#slots[slotNumber];
        yield slot.master.client;
        for (const replica of slot.replicas) {
            yield replica.client;
        }
    }
    #getSlotClient(slotNumber) {
        const slot = this.#slots[slotNumber];
        if (!slot.clientIterator) {
            slot.clientIterator = this.#slotClientIterator(slotNumber);
        }
        const { done, value } = slot.clientIterator.next();
        if (done) {
            slot.clientIterator = undefined;
            return this.#getSlotClient(slotNumber);
        }
        return value;
    }
    #randomClientIterator;
    #getRandomClient() {
        if (!this.#nodeByUrl.size) {
            throw new Error('Cluster is not connected');
        }
        if (!this.#randomClientIterator) {
            this.#randomClientIterator = this.#nodeByUrl.values();
        }
        const { done, value } = this.#randomClientIterator.next();
        if (done) {
            this.#randomClientIterator = undefined;
            return this.#getRandomClient();
        }
        return value.client;
    }
    getClient(firstKey, isReadonly) {
        if (!firstKey) {
            return this.#getRandomClient();
        }
        const slot = cluster_key_slot_1.default(firstKey);
        if (!isReadonly || !this.#options.useReplicas) {
            return this.getSlotMaster(slot).client;
        }
        return this.#getSlotClient(slot);
    }
    getMasters() {
        const masters = [];
        for (const node of this.#nodeByUrl.values()) {
            if (node.client.options?.readonly)
                continue;
            masters.push(node);
        }
        return masters;
    }
    getNodeByUrl(url) {
        return this.#nodeByUrl.get(url);
    }
    async disconnect() {
        await Promise.all([...this.#nodeByUrl.values()].map(({ client }) => client.disconnect()));
        this.#nodeByUrl.clear();
        this.#slots.splice(0);
    }
}
exports.default = RedisClusterSlots;
