"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const commands_1 = __importDefault(require("./commands"));
const cluster_slots_1 = __importDefault(require("./cluster-slots"));
const command_options_1 = require("./command-options");
class RedisCluster {
    static defineCommand(on, name, command) {
        on[name] = async function (...args) {
            const options = command_options_1.isCommandOptions(args[0]) && args[0], redisArgs = command.transformArguments(...(options ? args.slice(1) : args));
            return command.transformReply(await this.sendCommand(RedisCluster.#extractFirstKey(command, args, redisArgs), command.IS_READ_ONLY, redisArgs, options));
        };
    }
    static #extractFirstKey(commandOrScript, originalArgs, redisArgs) {
        if (commandOrScript.FIRST_KEY_INDEX === undefined) {
            return undefined;
        }
        else if (typeof commandOrScript.FIRST_KEY_INDEX === 'number') {
            return redisArgs[commandOrScript.FIRST_KEY_INDEX];
        }
        return commandOrScript.FIRST_KEY_INDEX(...originalArgs);
    }
    static create(options) {
        return new RedisCluster(options);
    }
    static commandOptions(options) {
        return command_options_1.commandOptions(options);
    }
    #options;
    #slots;
    constructor(options) {
        this.#options = options;
        this.#slots = new cluster_slots_1.default(options);
        this.#initiateModules();
        this.#initiateScripts();
    }
    #initiateModules() {
        if (!this.#options.modules)
            return;
        for (const m of this.#options.modules) {
            for (const [name, command] of Object.entries(m)) {
                RedisCluster.defineCommand(this, name, command);
            }
        }
    }
    #initiateScripts() {
        if (!this.#options.scripts)
            return;
        for (const [name, script] of Object.entries(this.#options.scripts)) {
            this[name] = async function (...args) {
                const options = command_options_1.isCommandOptions(args[0]) && args[0];
                return script.transformReply(await this.executeScript(script, args, script.transformArguments(...(options ? args.slice(1) : args)), options));
            };
        }
    }
    async connect() {
        return this.#slots.connect();
    }
    async sendCommand(firstKey, isReadonly, args, options, redirections = 0) {
        const client = this.#slots.getClient(firstKey, isReadonly);
        try {
            return await client.sendCommand(args, options);
        }
        catch (err) {
            const shouldRetry = await this.#handleCommandError(err, client, redirections);
            if (shouldRetry === true) {
                return this.sendCommand(firstKey, isReadonly, args, options, redirections + 1);
            }
            else if (shouldRetry) {
                return shouldRetry.sendCommand(args, options);
            }
            throw err;
        }
    }
    async executeScript(script, originalArgs, redisArgs, options, redirections = 0) {
        const client = this.#slots.getClient(RedisCluster.#extractFirstKey(script, originalArgs, redisArgs), script.IS_READ_ONLY);
        try {
            return await client.executeScript(script, redisArgs, options);
        }
        catch (err) {
            const shouldRetry = await this.#handleCommandError(err, client, redirections);
            if (shouldRetry === true) {
                return this.executeScript(script, originalArgs, redisArgs, options, redirections + 1);
            }
            else if (shouldRetry) {
                return shouldRetry.executeScript(script, redisArgs, options);
            }
            throw err;
        }
    }
    async #handleCommandError(err, client, redirections) {
        if (redirections > (this.#options.maxCommandRedirections ?? 16)) {
            throw err;
        }
        if (err.message.startsWith('ASK')) {
            const url = err.message.substring(err.message.lastIndexOf(' ') + 1);
            let node = this.#slots.getNodeByUrl(url);
            if (!node) {
                await this.#slots.discover(client);
                node = this.#slots.getNodeByUrl(url);
                if (!node) {
                    throw new Error(`Cannot find node ${url}`);
                }
            }
            await node.client.asking();
            return node.client;
        }
        else if (err.message.startsWith('MOVED')) {
            await this.#slots.discover(client);
            return client;
        }
        throw err;
    }
    getMasters() {
        return this.#slots.getMasters();
    }
    getSlotMaster(slot) {
        return this.#slots.getSlotMaster(slot);
    }
    disconnect() {
        return this.#slots.disconnect();
    }
}
exports.default = RedisCluster;
for (const [name, command] of Object.entries(commands_1.default)) {
    RedisCluster.defineCommand(RedisCluster.prototype, name, command);
}
